from sonusai.mixture.datatypes import ImpulseResponseFile
from sonusai.mixture.datatypes import NoiseFile
from sonusai.mixture.datatypes import SpectralMask
from sonusai.mixture.datatypes import TargetFile
from sonusai.mixture.datatypes import TruthParameter


def raw_load_config(name: str) -> dict:
    """Load YAML config file

    :param name: File name
    :return: Dictionary of config data
    """
    import yaml

    with open(file=name) as f:
        config = yaml.safe_load(f)

    return config


def get_default_config() -> dict:
    """Load default SonusAI config

    :return: Dictionary of default config data
    """
    from .constants import DEFAULT_CONFIG

    try:
        return raw_load_config(DEFAULT_CONFIG)
    except Exception as e:
        raise OSError(f"Error loading default config: {e}") from e


def load_config(name: str) -> dict:
    """Load SonusAI default config and update with given location (performing SonusAI variable substitution)

    :param name: Directory containing mixture database
    :return: Dictionary of config data
    """
    from os.path import join

    return update_config_from_file(filename=join(name, "config.yml"), given_config=get_default_config())


def update_config_from_file(filename: str, given_config: dict) -> dict:
    """Update the given config with the config in the specified YAML file

    :param filename: File name
    :param given_config: Config dictionary to update
    :return: Updated config dictionary
    """
    from copy import deepcopy

    from .constants import REQUIRED_CONFIGS
    from .constants import VALID_CONFIGS
    from .constants import VALID_NOISE_MIX_MODES

    updated_config = deepcopy(given_config)

    try:
        file_config = raw_load_config(filename)
    except Exception as e:
        raise OSError(f"Error loading config from {filename}: {e}") from e

    # Check for unrecognized keys
    for key in file_config:
        if key not in VALID_CONFIGS:
            nice_list = "\n".join([f"  {item}" for item in VALID_CONFIGS])
            raise AttributeError(
                f"Invalid config parameter in {filename}: {key}.\nValid config parameters are:\n{nice_list}"
            )

    # Use default config as base and overwrite with given config keys as found
    for key in updated_config:
        if key in file_config:
            updated_config[key] = file_config[key]

    # Check for required keys
    for key in REQUIRED_CONFIGS:
        if key not in updated_config:
            raise AttributeError(f"{filename} is missing required '{key}'")

    # Validate special cases
    validate_truth_configs(updated_config)
    validate_asr_configs(updated_config)

    # Check for non-empty spectral masks
    if len(updated_config["spectral_masks"]) == 0:
        updated_config["spectral_masks"] = given_config["spectral_masks"]

    # Check for valid noise_mix_mode
    if updated_config["noise_mix_mode"] not in VALID_NOISE_MIX_MODES:
        nice_list = "\n".join([f"  {item}" for item in VALID_NOISE_MIX_MODES])
        raise ValueError(f"{filename} contains invalid noise_mix_mode.\nValid noise mix modes are:\n{nice_list}")

    return updated_config


def validate_truth_configs(given: dict) -> None:
    """Validate fields in given 'truth_configs'

    :param given: The dictionary of given config
    """
    from copy import deepcopy

    from sonusai.mixture import truth_functions

    from .constants import REQUIRED_TRUTH_CONFIGS

    if "truth_configs" not in given:
        raise AttributeError("config is missing required 'truth_configs'")

    truth_configs = given["truth_configs"]
    if len(truth_configs) == 0:
        raise ValueError("'truth_configs' in config is empty")

    for name, truth_config in truth_configs.items():
        for key in REQUIRED_TRUTH_CONFIGS:
            if key not in truth_config:
                raise AttributeError(f"'{name}' in truth_configs is missing required '{key}'")

        optional_config = deepcopy(truth_config)
        for key in REQUIRED_TRUTH_CONFIGS:
            del optional_config[key]

        getattr(truth_functions, truth_config["function"] + "_validate")(optional_config)


def validate_asr_configs(given: dict) -> None:
    """Validate fields in given 'asr_config'

    :param given: The dictionary of given config
    """
    from sonusai.utils import validate_asr

    from .constants import REQUIRED_ASR_CONFIGS

    if "asr_configs" not in given:
        raise AttributeError("config is missing required 'asr_configs'")

    asr_configs = given["asr_configs"]

    for name, asr_config in asr_configs.items():
        for key in REQUIRED_ASR_CONFIGS:
            if key not in asr_config:
                raise AttributeError(f"'{name}' in asr_configs is missing required '{key}'")

        engine = asr_config["engine"]
        config = {x: asr_config[x] for x in asr_config if x != "engine"}
        validate_asr(engine, **config)


def get_hierarchical_config_files(root: str, leaf: str) -> list[str]:
    """Get a hierarchical list of config files in the given leaf of the given root

    :param root: Root of the hierarchy
    :param leaf: Leaf under the root
    :return: List of config files found in the hierarchy
    """
    import os
    from pathlib import Path

    config_file = "config.yml"

    root_path = Path(os.path.abspath(root))
    if not root_path.is_dir():
        raise OSError(f"Given root, {root_path}, is not a directory.")

    leaf_path = Path(os.path.abspath(leaf))
    if not leaf_path.is_dir():
        raise OSError(f"Given leaf, {leaf_path}, is not a directory.")

    common = os.path.commonpath((root_path, leaf_path))
    if os.path.normpath(common) != os.path.normpath(root_path):
        raise OSError(f"Given leaf, {leaf_path}, is not in the hierarchy of the given root, {root_path}")

    top_config_file = os.path.join(root_path, config_file)
    if not Path(top_config_file).is_file():
        raise OSError(f"Could not find {top_config_file}")

    current = leaf_path
    config_files = []
    while current != root_path:
        local_config_file = Path(os.path.join(current, config_file))
        if local_config_file.is_file():
            config_files.append(str(local_config_file))
        current = current.parent

    config_files.append(top_config_file)
    return list(reversed(config_files))


def update_config_from_hierarchy(root: str, leaf: str, config: dict) -> dict:
    """Update the given config using the hierarchical config files in the given leaf of the given root

    :param root: Root of the hierarchy
    :param leaf: Leaf under the root
    :param config: Config to update
    :return: Updated config
    """
    from copy import deepcopy

    new_config = deepcopy(config)
    config_files = get_hierarchical_config_files(root=root, leaf=leaf)
    for config_file in config_files:
        new_config = update_config_from_file(filename=config_file, given_config=new_config)

    return new_config


def get_target_files(config: dict, show_progress: bool = False) -> list[TargetFile]:
    """Get the list of target files from a config

    :param config: Config dictionary
    :param show_progress: Show progress bar
    :return: List of target files
    """
    from itertools import chain

    from sonusai.utils import dataclass_from_dict
    from sonusai.utils import par_track
    from sonusai.utils import track

    from .datatypes import TargetFile

    class_indices = config["class_indices"]
    if not isinstance(class_indices, list):
        class_indices = [class_indices]

    target_files = list(
        chain.from_iterable(
            [
                append_target_files(
                    entry=entry,
                    class_indices=class_indices,
                    truth_configs=config["truth_configs"],
                    level_type=config["target_level_type"],
                )
                for entry in config["targets"]
            ]
        )
    )

    progress = track(total=len(target_files), disable=not show_progress)
    target_files = par_track(_get_num_samples, target_files, progress=progress)
    progress.close()

    num_classes = config["num_classes"]
    for target_file in target_files:
        if any(class_index < 0 for class_index in target_file["class_indices"]):
            raise ValueError("class indices must contain only positive elements")

        if any(class_index > num_classes for class_index in target_file["class_indices"]):
            raise ValueError(f"class index elements must not be greater than {num_classes}")

    return dataclass_from_dict(list[TargetFile], target_files)


def append_target_files(
    entry: dict | str,
    class_indices: list[int],
    truth_configs: dict,
    level_type: str,
    tokens: dict | None = None,
) -> list[dict]:
    """Process target files list and append as needed

    :param entry: Target file entry to append to the list
    :param class_indices: Class indices
    :param truth_configs: Truth configs
    :param level_type: Target level type
    :param tokens: Tokens used for variable expansion
    :return: List of target files
    """
    from copy import deepcopy
    from glob import glob
    from os import listdir
    from os.path import dirname
    from os.path import isabs
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    from sonusai.utils import dataclass_from_dict

    from .audio import validate_input_file
    from .constants import REQUIRED_TRUTH_CONFIGS
    from .datatypes import TruthConfig
    from .tokenized_shell_vars import tokenized_expand
    from .tokenized_shell_vars import tokenized_replace

    if tokens is None:
        tokens = {}

    truth_configs_merged = deepcopy(truth_configs)
    if isinstance(entry, dict):
        if "name" in entry:
            in_name = entry["name"]
        else:
            raise AttributeError("Target list contained record without name")

        if "class_indices" in entry:
            if isinstance(entry["class_indices"], list):
                class_indices = entry["class_indices"]
            else:
                class_indices = [entry["class_indices"]]

        truth_configs_override = entry.get("truth_configs", {})
        for key in truth_configs_override:
            if key not in truth_configs:
                raise AttributeError(
                    f"Truth config '{key}' override specified for {entry['name']} is not defined at top level"
                )
            if key in truth_configs_override:
                truth_configs_merged[key] |= truth_configs_override[key]
        level_type = entry.get("level_type", level_type)
    else:
        in_name = entry

    in_name, new_tokens = tokenized_expand(in_name)
    tokens.update(new_tokens)
    names = sorted(glob(in_name))
    if not names:
        raise OSError(f"Could not find {in_name}. Make sure path exists")

    target_files: list[dict] = []
    for name in names:
        ext = splitext(name)[1].lower()
        dir_name = dirname(name)
        if isdir(name):
            for file in listdir(name):
                child = file
                if not isabs(child):
                    child = join(dir_name, child)
                target_files.extend(
                    append_target_files(
                        entry=child,
                        class_indices=class_indices,
                        truth_configs=truth_configs_merged,
                        level_type=level_type,
                        tokens=tokens,
                    )
                )
        else:
            try:
                if ext == ".txt":
                    with open(file=name) as txt_file:
                        for line in txt_file:
                            # strip comments
                            child = line.partition("#")[0]
                            child = child.rstrip()
                            if child:
                                child, new_tokens = tokenized_expand(child)
                                tokens.update(new_tokens)
                                if not isabs(child):
                                    child = join(dir_name, child)
                                target_files.extend(
                                    append_target_files(
                                        entry=child,
                                        class_indices=class_indices,
                                        truth_configs=truth_configs_merged,
                                        level_type=level_type,
                                        tokens=tokens,
                                    )
                                )
                elif ext == ".yml":
                    try:
                        yml_config = raw_load_config(name)

                        if "targets" in yml_config:
                            for record in yml_config["targets"]:
                                target_files.extend(
                                    append_target_files(
                                        entry=record,
                                        class_indices=class_indices,
                                        truth_configs=truth_configs_merged,
                                        level_type=level_type,
                                        tokens=tokens,
                                    )
                                )
                    except Exception as e:
                        raise OSError(f"Error processing {name}: {e}") from e
                else:
                    validate_input_file(name)
                    target_file: dict = {
                        "expanded_name": name,
                        "name": tokenized_replace(name, tokens),
                        "class_indices": class_indices,
                        "level_type": level_type,
                        "truth_configs": {},
                    }
                    if len(truth_configs_merged) > 0:
                        for tc_key, tc_value in truth_configs_merged.items():
                            config = deepcopy(tc_value)
                            truth_config: dict = {}
                            for key in REQUIRED_TRUTH_CONFIGS:
                                truth_config[key] = config[key]
                                del config[key]
                            truth_config["config"] = config
                            target_file["truth_configs"][tc_key] = dataclass_from_dict(TruthConfig, truth_config)
                        for tc_key in target_file["truth_configs"]:
                            if (
                                "function" in truth_configs_merged[tc_key]
                                and truth_configs_merged[tc_key]["function"] == "file"
                            ):
                                truth_configs_merged[tc_key]["file"] = splitext(target_file["name"])[0] + ".h5"
                    target_files.append(target_file)
            except Exception as e:
                raise OSError(f"Error processing {name}: {e}") from e

    return target_files


def get_noise_files(config: dict, show_progress: bool = False) -> list[NoiseFile]:
    """Get the list of noise files from a config

    :param config: Config dictionary
    :param show_progress: Show progress bar
    :return: List of noise file
    """
    from itertools import chain

    from sonusai.utils import dataclass_from_dict
    from sonusai.utils import par_track
    from sonusai.utils import track

    from .datatypes import NoiseFile

    noise_files = list(chain.from_iterable([append_noise_files(entry=entry) for entry in config["noises"]]))

    progress = track(total=len(noise_files), disable=not show_progress)
    noise_files = par_track(_get_num_samples, noise_files, progress=progress)
    progress.close()

    return dataclass_from_dict(list[NoiseFile], noise_files)


def append_noise_files(entry: dict | str, tokens: dict | None = None) -> list[dict]:
    """Process noise files list and append as needed

    :param entry: Noise file entry to append to the list
    :param tokens: Tokens used for variable expansion
    :return: List of noise files
    """
    from glob import glob
    from os import listdir
    from os.path import dirname
    from os.path import isabs
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    from .audio import validate_input_file
    from .tokenized_shell_vars import tokenized_expand
    from .tokenized_shell_vars import tokenized_replace

    if tokens is None:
        tokens = {}

    if isinstance(entry, dict):
        if "name" in entry:
            in_name = entry["name"]
        else:
            raise AttributeError("Noise list contained record without name")
    else:
        in_name = entry

    in_name, new_tokens = tokenized_expand(in_name)
    tokens.update(new_tokens)
    names = sorted(glob(in_name))
    if not names:
        raise OSError(f"Could not find {in_name}. Make sure path exists")

    noise_files: list[dict] = []
    for name in names:
        ext = splitext(name)[1].lower()
        dir_name = dirname(name)
        if isdir(name):
            for file in listdir(name):
                child = file
                if not isabs(child):
                    child = join(dir_name, child)
                noise_files.extend(append_noise_files(entry=child, tokens=tokens))
        else:
            try:
                if ext == ".txt":
                    with open(file=name) as txt_file:
                        for line in txt_file:
                            # strip comments
                            child = line.partition("#")[0]
                            child = child.rstrip()
                            if child:
                                child, new_tokens = tokenized_expand(child)
                                tokens.update(new_tokens)
                                if not isabs(child):
                                    child = join(dir_name, child)
                                noise_files.extend(append_noise_files(entry=child, tokens=tokens))
                elif ext == ".yml":
                    try:
                        yml_config = raw_load_config(name)

                        if "noises" in yml_config:
                            for record in yml_config["noises"]:
                                noise_files.extend(append_noise_files(entry=record, tokens=tokens))
                    except Exception as e:
                        raise OSError(f"Error processing {name}: {e}") from e
                else:
                    validate_input_file(name)
                    noise_file: dict = {
                        "expanded_name": name,
                        "name": tokenized_replace(name, tokens),
                    }
                    noise_files.append(noise_file)
            except Exception as e:
                raise OSError(f"Error processing {name}: {e}") from e

    return noise_files


def get_impulse_response_files(config: dict) -> list[ImpulseResponseFile]:
    """Get the list of impulse response files from a config

    :param config: Config dictionary
    :return: List of impulse response files
    """
    from itertools import chain

    return list(
        chain.from_iterable(
            [
                append_impulse_response_files(entry=ImpulseResponseFile(entry["name"], entry.get("tags", [])))
                for entry in config["impulse_responses"]
            ]
        )
    )


def append_impulse_response_files(entry: ImpulseResponseFile, tokens: dict | None = None) -> list[ImpulseResponseFile]:
    """Process impulse response files list and append as needed

    :param entry: Impulse response file entry to append to the list
    :param tokens: Tokens used for variable expansion
    :return: List of impulse response files
    """
    from glob import glob
    from os import listdir
    from os.path import dirname
    from os.path import isabs
    from os.path import isdir
    from os.path import join
    from os.path import splitext

    from .audio import validate_input_file
    from .tokenized_shell_vars import tokenized_expand
    from .tokenized_shell_vars import tokenized_replace

    if tokens is None:
        tokens = {}

    in_name, new_tokens = tokenized_expand(entry.file)
    tokens.update(new_tokens)
    names = sorted(glob(in_name))
    if not names:
        raise OSError(f"Could not find {in_name}. Make sure path exists")

    impulse_response_files: list[ImpulseResponseFile] = []
    for name in names:
        ext = splitext(name)[1].lower()
        dir_name = dirname(name)
        if isdir(name):
            for file in listdir(name):
                if not isabs(file):
                    file = join(dir_name, file)
                child = ImpulseResponseFile(file, entry.tags)
                impulse_response_files.extend(append_impulse_response_files(entry=child, tokens=tokens))
        else:
            try:
                if ext == ".txt":
                    with open(file=name) as txt_file:
                        for line in txt_file:
                            # strip comments
                            file = line.partition("#")[0]
                            file = file.rstrip()
                            if file:
                                file, new_tokens = tokenized_expand(file)
                                tokens.update(new_tokens)
                                if not isabs(file):
                                    file = join(dir_name, file)
                                child = ImpulseResponseFile(file, entry.tags)
                                impulse_response_files.extend(append_impulse_response_files(entry=child, tokens=tokens))
                elif ext == ".yml":
                    try:
                        yml_config = raw_load_config(name)

                        if "impulse_responses" in yml_config:
                            for record in yml_config["impulse_responses"]:
                                impulse_response_files.extend(
                                    append_impulse_response_files(entry=record, tokens=tokens)
                                )
                    except Exception as e:
                        raise OSError(f"Error processing {name}: {e}") from e
                else:
                    validate_input_file(name)
                    impulse_response_files.append(ImpulseResponseFile(tokenized_replace(name, tokens), entry.tags))
            except Exception as e:
                raise OSError(f"Error processing {name}: {e}") from e

    return impulse_response_files


def get_spectral_masks(config: dict) -> list[SpectralMask]:
    """Get the list of spectral masks from a config

    :param config: Config dictionary
    :return: List of spectral masks
    """
    from sonusai.utils import dataclass_from_dict

    try:
        return dataclass_from_dict(list[SpectralMask], config["spectral_masks"])
    except Exception as e:
        raise ValueError(f"Error in spectral_masks: {e}") from e


def get_truth_parameters(config: dict) -> list[TruthParameter]:
    """Get the list of truth parameters from a config

    :param config: Config dictionary
    :return: List of truth parameters
    """
    from copy import deepcopy

    from sonusai.mixture import truth_functions

    from .constants import REQUIRED_TRUTH_CONFIGS
    from .datatypes import TruthParameter

    truth_parameters: list[TruthParameter] = []
    for name, truth_config in config["truth_configs"].items():
        optional_config = deepcopy(truth_config)
        for key in REQUIRED_TRUTH_CONFIGS:
            del optional_config[key]

        parameters = getattr(truth_functions, truth_config["function"] + "_parameters")(
            config["feature"],
            config["num_classes"],
            optional_config,
        )
        truth_parameters.append(TruthParameter(name, parameters))

    return truth_parameters


def _get_num_samples(entry: dict) -> dict:
    from .audio import get_num_samples

    entry["samples"] = get_num_samples(entry["expanded_name"])
    del entry["expanded_name"]
    return entry



import re
from importlib.resources import as_file
from importlib.resources import files

REQUIRED_CONFIGS = [
    "asr_configs",
    "class_balancing",
    "class_balancing_augmentation",
    "class_indices",
    "class_labels",
    "class_weights_threshold",
    "feature",
    "impulse_responses",
    "noise_augmentations",
    "noise_mix_mode",
    "noises",
    "num_classes",
    "random_snrs",
    "seed",
    "snrs",
    "spectral_masks",
    "target_augmentations",
    "target_level_type",
    "targets",
    "truth_configs",
]
OPTIONAL_CONFIGS: list[str] = []
VALID_CONFIGS = REQUIRED_CONFIGS + OPTIONAL_CONFIGS
REQUIRED_TRUTH_CONFIGS = ["function", "stride_reduction"]
REQUIRED_ASR_CONFIGS = ["engine"]
VALID_AUGMENTATIONS = [
    "normalize",
    "gain",
    "pitch",
    "tempo",
    "eq1",
    "eq2",
    "eq3",
    "lpf",
    "ir",
    "mixup",
]
VALID_NOISE_MIX_MODES = ["exhaustive", "non-exhaustive", "non-combinatorial"]
RAND_PATTERN = re.compile(r"rand\(([-+]?(\d+(\.\d*)?|\.\d+)),\s*([-+]?(\d+(\.\d*)?|\.\d+))\)")
SAMPLE_RATE = 16000
BIT_DEPTH = 32
ENCODING = "floating-point"
CHANNEL_COUNT = 1
SAMPLE_BYTES = BIT_DEPTH // 8
FLOAT_BYTES = 4
MIXDB_VERSION = 2

with as_file(files("sonusai.data").joinpath("whitenoise.wav")) as path:
    DEFAULT_NOISE = str(path)

with as_file(files("sonusai.data").joinpath("genmixdb.yml")) as path:
    DEFAULT_CONFIG = str(path)

with as_file(files("sonusai.data").joinpath("speech_ma01_01.wav")) as path:
    DEFAULT_SPEECH = str(path)



from pathlib import Path

import numpy as np
from sox import Transformer as SoxTransformer

from sonusai.mixture.datatypes import AudioT
from sonusai.mixture.datatypes import ImpulseResponseData


def read_impulse_response(name: str | Path) -> ImpulseResponseData:
    """Read impulse response data using SoX

    :param name: File name
    :return: ImpulseResponseData object
    """
    from scipy.io import wavfile

    from .datatypes import ImpulseResponseData
    from .tokenized_shell_vars import tokenized_expand

    expanded_name, _ = tokenized_expand(name)

    # Read impulse response data from audio file
    try:
        sample_rate, data = wavfile.read(expanded_name)
    except Exception as e:
        if name != expanded_name:
            raise OSError(f"Error reading {name} (expanded: {expanded_name}): {e}") from e
        else:
            raise OSError(f"Error reading {name}: {e}") from e

    data = data.astype(np.float32)
    offset = np.argmax(data)
    data = data[offset:]
    data = data / np.linalg.norm(data)

    return ImpulseResponseData(name=str(name), sample_rate=sample_rate, data=data)


def read_audio(name: str | Path) -> AudioT:
    """Read audio data from a file using SoX

    :param name: File name
    :return: Array of time domain audio data
    """
    from typing import Any

    from sox.core import sox

    from .constants import BIT_DEPTH
    from .constants import CHANNEL_COUNT
    from .constants import ENCODING
    from .constants import SAMPLE_RATE
    from .tokenized_shell_vars import tokenized_expand

    def encode_output(buffer: Any) -> np.ndarray:
        from .constants import BIT_DEPTH
        from .constants import ENCODING

        if BIT_DEPTH == 8:
            return np.frombuffer(buffer, dtype=np.int8)

        if BIT_DEPTH == 16:
            return np.frombuffer(buffer, dtype=np.int16)

        if BIT_DEPTH == 24:
            return np.frombuffer(buffer, dtype=np.int32)

        if BIT_DEPTH == 32:
            if ENCODING == "floating-point":
                return np.frombuffer(buffer, dtype=np.float32)
            return np.frombuffer(buffer, dtype=np.int32)

        if BIT_DEPTH == 64:
            return np.frombuffer(buffer, dtype=np.float64)

        raise ValueError(f"Invalid BIT_DEPTH {BIT_DEPTH}")

    expanded_name, _ = tokenized_expand(name)

    try:
        # Read in and convert to desired format
        # NOTE: pysox format transformations do not handle encoding properly; need to use direct call to sox instead
        args = [
            "-D",
            "-G",
            expanded_name,
            "-t",
            "raw",
            "-r",
            str(SAMPLE_RATE),
            "-b",
            str(BIT_DEPTH),
            "-c",
            str(CHANNEL_COUNT),
            "-e",
            ENCODING,
            "-",
            "remix",
            "1",
        ]
        status, out, err = sox(args, None, False)
        if status != 0:
            raise RuntimeError(f"sox stdout: {out}\nsox stderr: {err}")  # noqa: TRY301

        return encode_output(out)

    except Exception as e:
        if name != expanded_name:
            raise OSError(f"Error reading {name} (expanded: {expanded_name}):\n{e}") from e
        else:
            raise OSError(f"Error reading {name}:\n{e}") from e


class Transformer(SoxTransformer):
    """Override certain sox.Transformer methods"""

    def fir(self, coefficients):
        """Use SoX's FFT convolution engine with given FIR filter coefficients.

        The SonusAI override allows coefficients to be either a list of numbers
        or a string containing a text file with the coefficients.

        Parameters
        ----------
        coefficients : list or str
            fir filter coefficients

        """
        from sox.core import is_number

        if not isinstance(coefficients, list) and not isinstance(coefficients, str):
            raise TypeError("coefficients must be a list or a str.")

        if isinstance(coefficients, list) and not all(is_number(c) for c in coefficients):
            raise TypeError("coefficients list must be numbers.")

        effect_args = ["fir"]
        if isinstance(coefficients, list):
            effect_args.extend([f"{c:f}" for c in coefficients])
        else:
            effect_args.append(coefficients)

        self.effects.extend(effect_args)
        self.effects_log.append("fir")

        return self

    def tempo(self, factor, audio_type=None, quick=False):
        """Time stretch audio without changing pitch.

        This effect uses the WSOLA algorithm. The audio is chopped up into
        segments which are then shifted in the time domain and overlapped
        (cross-faded) at points where their waveforms are most similar as
        determined by measurement of least squares.

        The SonusAI override does not generate a warning for small factors.
        The sox.Transformer's implementation of stretch does not invert
        the factor even though it says that it does; this invalidates the
        factor size check and produces the wrong result.

        Parameters
        ----------
        factor : float
            The ratio of new tempo to the old tempo.
            For ex. 1.1 speeds up the tempo by 10%; 0.9 slows it down by 10%.
        audio_type : str
            Type of audio, which optimizes algorithm parameters. One of:
             * m : Music,
             * s : Speech,
             * l : Linear (useful when factor is close to 1),
        quick : bool, default=False
            If True, this effect will run faster but with lower sound quality.

        See Also
        --------
        stretch, speed, pitch

        """
        from sox.core import is_number
        from sox.log import logger

        if not is_number(factor) or factor <= 0:
            raise ValueError("factor must be a positive number")

        if factor < 0.5 or factor > 2:
            logger.warning("Using an extreme time stretching factor. Quality of results will be poor")

        if audio_type not in [None, "m", "s", "l"]:
            raise ValueError("audio_type must be one of None, 'm', 's', or 'l'.")

        if not isinstance(quick, bool):
            raise TypeError("quick must be a boolean")

        effect_args = ["tempo"]

        if quick:
            effect_args.append("-q")

        if audio_type is not None:
            effect_args.append(f"-{audio_type}")

        effect_args.append(f"{factor:f}")

        self.effects.extend(effect_args)
        self.effects_log.append("tempo")

        return self

    def build(  # pyright: ignore [reportIncompatibleMethodOverride]
        self,
        input_filepath: str | Path | None = None,
        output_filepath: str | Path | None = None,
        input_array: np.ndarray | None = None,
        sample_rate_in: float | None = None,
        extra_args: list[str] | None = None,
        return_output: bool = False,
    ) -> tuple[bool, str | None, str | None]:
        """Given an input file or array, creates an output_file on disk by
        executing the current set of commands. This function returns True on
        success. If return_output is True, this function returns a triple of
        (status, out, err), giving the success state, along with stdout and
        stderr returned by sox.

        Parameters
        ----------
        input_filepath : str or None
            Either path to input audio file or None for array input.
        output_filepath : str
            Path to desired output file. If a file already exists at
            the given path, the file will be overwritten.
            If '-n', no file is created.
        input_array : np.ndarray or None
            An np.ndarray of an waveform with shape (n_samples, n_channels).
            sample_rate_in must also be provided.
            If None, input_filepath must be specified.
        sample_rate_in : int
            Sample rate of input_array.
            This argument is ignored if input_array is None.
        extra_args : list or None, default=None
            If a list is given, these additional arguments are passed to SoX
            at the end of the list of effects.
            Don't use this argument unless you know exactly what you're doing!
        return_output : bool, default=False
            If True, returns the status and information sent to stderr and
            stdout as a tuple (status, stdout, stderr).
            If output_filepath is None, return_output=True by default.
            If False, returns True on success.

        Returns
        -------
        status : bool
            True on success.
        out : str (optional)
            This is not returned unless return_output is True.
            When returned, captures the stdout produced by sox.
        err : str (optional)
            This is not returned unless return_output is True.
            When returned, captures the stderr produced by sox.

        Examples
        --------
        > import numpy as np
        > import sox
        > tfm = sox.Transformer()
        > sample_rate = 44100
        > y = np.sin(2 * np.pi * 440.0 * np.arange(sample_rate * 1.0) / sample_rate)

        file in, file out - basic usage

        > status = tfm.build('path/to/input.wav', 'path/to/output.mp3')

        file in, file out - equivalent usage

        > status = tfm.build(
                input_filepath='path/to/input.wav',
                output_filepath='path/to/output.mp3'
            )

        array in, file out

        > status = tfm.build(
                input_array=y, sample_rate_in=sample_rate,
                output_filepath='path/to/output.mp3'
            )

        """
        from sox import file_info
        from sox.core import SoxError
        from sox.core import sox
        from sox.log import logger

        input_format, input_filepath = self._parse_inputs(input_filepath, input_array, sample_rate_in)

        if output_filepath is None:
            raise ValueError("output_filepath is not specified!")

        # set output parameters
        if input_filepath == output_filepath:
            raise ValueError("input_filepath must be different from output_filepath.")
        file_info.validate_output_file(output_filepath)

        args = []
        args.extend(self.globals)
        args.extend(self._input_format_args(input_format))
        args.append(input_filepath)
        args.extend(self._output_format_args(self.output_format))
        args.append(output_filepath)
        args.extend(self.effects)

        if extra_args is not None:
            if not isinstance(extra_args, list):
                raise ValueError("extra_args must be a list.")
            args.extend(extra_args)

        status, out, err = sox(args, input_array, True)
        if status != 0:
            raise SoxError(f"Stdout: {out}\nStderr: {err}")

        logger.info("Created %s with effects: %s", output_filepath, " ".join(self.effects_log))

        if return_output:
            return status, out, err  # pyright: ignore [reportReturnType]

        return True, None, None

    def build_array(  # pyright: ignore [reportIncompatibleMethodOverride]
        self,
        input_filepath: str | Path | None = None,
        input_array: np.ndarray | None = None,
        sample_rate_in: int | None = None,
        extra_args: list[str] | None = None,
    ) -> np.ndarray:
        """Given an input file or array, returns the output as a numpy array
        by executing the current set of commands. By default, the array will
        have the same sample rate as the input file unless otherwise specified
        using set_output_format. Functions such as channels and convert
        will be ignored!

        The SonusAI override does not generate a warning for rate transforms.

        Parameters
        ----------
        input_filepath : str, Path or None
            Either path to input audio file or None.
        input_array : np.ndarray or None
            A np.ndarray of a waveform with shape (n_samples, n_channels).
            If this argument is passed, sample_rate_in must also be provided.
            If None, input_filepath must be specified.
        sample_rate_in : int
            Sample rate of input_array.
            This argument is ignored if input_array is None.
        extra_args : list or None, default=None
            If a list is given, these additional arguments are passed to SoX
            at the end of the list of effects.
            Don't use this argument unless you know exactly what you're doing!

        Returns
        -------
        output_array : np.ndarray
            Output audio as a numpy array

        Examples
        --------

        > import numpy as np
        > import sox
        > tfm = sox.Transformer()
        > sample_rate = 44100
        > y = np.sin(2 * np.pi * 440.0 * np.arange(sample_rate * 1.0) / sample_rate)

        file in, array out

        > output_array = tfm.build(input_filepath='path/to/input.wav')

        array in, array out

        > output_array = tfm.build(input_array=y, sample_rate_in=sample_rate)

        specifying the output sample rate

        > tfm.set_output_format(rate=8000)
        > output_array = tfm.build(input_array=y, sample_rate_in=sample_rate)

        if an effect changes the number of channels, you must explicitly
        specify the number of output channels

        > tfm.remix(remix_dictionary={1: [1], 2: [1], 3: [1]})
        > tfm.set_output_format(channels=3)
        > output_array = tfm.build(input_array=y, sample_rate_in=sample_rate)


        """
        from sox.core import SoxError
        from sox.core import sox
        from sox.log import logger
        from sox.transform import ENCODINGS_MAPPING

        input_format, input_filepath = self._parse_inputs(input_filepath, input_array, sample_rate_in)

        # check if any of the below commands are part of the effects chain
        ignored_commands = ["channels", "convert"]
        if set(ignored_commands) & set(self.effects_log):
            logger.warning(
                "When outputting to an array, channels and convert "
                + "effects may be ignored. Use set_output_format() to "
                + "specify output formats."
            )

        output_filepath = "-"

        if input_format.get("file_type") is None:
            encoding_out = np.int16
        else:
            encoding_out = next(k for k, v in ENCODINGS_MAPPING.items() if input_format["file_type"] == v)

        n_bits = np.dtype(encoding_out).itemsize * 8

        output_format = {
            "file_type": "raw",
            "rate": sample_rate_in,
            "bits": n_bits,
            "channels": input_format["channels"],
            "encoding": None,
            "comments": None,
            "append_comments": True,
        }

        if self.output_format.get("rate") is not None:
            output_format["rate"] = self.output_format["rate"]

        if self.output_format.get("channels") is not None:
            output_format["channels"] = self.output_format["channels"]

        if self.output_format.get("bits") is not None:
            n_bits = self.output_format["bits"]
            output_format["bits"] = n_bits

        match n_bits:
            case 8:
                encoding_out = np.int8  # type: ignore[assignment]
            case 16:
                encoding_out = np.int16
            case 32:
                encoding_out = np.float32  # type: ignore[assignment]
            case 64:
                encoding_out = np.float64  # type: ignore[assignment]
            case _:
                raise ValueError(f"invalid n_bits {n_bits}")

        args = []
        args.extend(self.globals)
        args.extend(self._input_format_args(input_format))
        args.append(input_filepath)
        args.extend(self._output_format_args(output_format))
        args.append(output_filepath)
        args.extend(self.effects)

        if extra_args is not None:
            if not isinstance(extra_args, list):
                raise ValueError("extra_args must be a list.")
            args.extend(extra_args)

        status, out, err = sox(args, input_array, False)
        if status != 0:
            raise SoxError(f"Stdout: {out}\nStderr: {err}")

        out = np.frombuffer(out, dtype=encoding_out)  # pyright: ignore [reportArgumentType, reportCallIssue]
        if output_format["channels"] > 1:
            out = out.reshape(
                (output_format["channels"], int(len(out) / output_format["channels"])),
                order="F",
            ).T
        logger.info("Created array with effects: %s", " ".join(self.effects_log))

        return out



from sonusai.mixture import MixtureDatabase
from sonusai.mixture import Truth


def truth_function(mixdb: MixtureDatabase, m_id: int) -> list[Truth]:
    from sonusai.mixture import TruthDict
    from sonusai.mixture import truth_functions

    result: list[Truth] = []
    for target_index in range(len(mixdb.mixture(m_id).targets)):
        truth: TruthDict = {}
        target_file = mixdb.target_file(mixdb.mixture(m_id).targets[target_index].file_id)
        for name, config in target_file.truth_configs.items():
            try:
                truth[name] = getattr(truth_functions, config.function)(mixdb, m_id, target_index, config.config)
            except AttributeError as e:
                raise AttributeError(f"Unsupported truth function: {config.function}") from e
            except Exception as e:
                raise RuntimeError(f"Error in truth function '{config.function}': {e}") from e

        result.append(truth)

    return result


def get_truth_indices_for_mixid(mixdb: MixtureDatabase, mixid: int) -> list[int]:
    """Get a list of truth indices for a given mixid."""
    indices: list[int] = []
    for target_id in [target.file_id for target in mixdb.mixture(mixid).targets]:
        indices.append(*mixdb.target_file(target_id).class_indices)

    return sorted(set(indices))


def truth_stride_reduction(truth: Truth, function: str) -> Truth:
    """Reduce stride dimension of truth.

    :param truth: Truth data [frames, stride, truth_parameters]
    :param function: Truth stride reduction function name
    :return: Stride reduced truth data [frames, stride or 1, truth_parameters]
    """
    import numpy as np

    if truth.ndim != 3:
        raise ValueError("Invalid truth shape")

    if function == "none":
        return truth

    if function == "max":
        return np.max(truth, axis=1, keepdims=True)

    if function == "mean":
        return np.mean(truth, axis=1, keepdims=True)

    if function == "first":
        return truth[:, 0, :].reshape((truth.shape[0], 1, truth.shape[2]))

    raise ValueError(f"Invalid truth stride reduction function: {function}")
